@@ -57,24 +57,6 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 int debug_ocl_icd_mask=0;
 
-typedef cl_uint cl_layer_info;
-typedef cl_uint cl_layer_api_version;
-#define CL_LAYER_API_VERSION 0x4240
-#define CL_LAYER_API_VERSION_100 100
-
-typedef __typeof__(clGetPlatformInfo) *clGetPlatformInfo_fn;
-CL_API_ENTRY typedef cl_int (CL_API_CALL *clGetLayerInfo_fn)(
-    cl_layer_info  param_name,
-    size_t         param_value_size,
-    void          *param_value,
-    size_t        *param_value_size_ret);
-
-CL_API_ENTRY typedef cl_int (CL_API_CALL *clInitLayer_fn)(
-    cl_uint                         num_entries,
-    const struct _cl_icd_dispatch  *target_dispatch,
-    cl_uint                        *num_entries_out,
-    const struct _cl_icd_dispatch **layer_dispatch);
-
 static inline void dump_vendor_icd(const char* info, const struct vendor_icd *v) {
   debug(D_DUMP, "%s %p={ num=%i, handle=%p, f=%p}\n", info,
 	v, v->num_platforms, v->dl_handle, v->ext_fn_ptr);
@@ -177,30 +159,58 @@ static char* _clerror2string (cl_int error) {
 #endif
 }
 
-static inline int _string_end_with_icd(const char* str) {
+static inline int _string_end_with(const char* str, const char* suffix) {
   size_t len = strlen(str);
-  if( len<5 || strcmp(str + len - 4, ".icd" ) != 0 ) {
+  size_t len_suff = strlen(suffix);
+  if (len < len_suff + 1 || strcmp(str + len - len_suff, suffix) != 0) {
     return 0;
   }
   return 1;
 }
 
+#define ICD_EXTENSION ".icd"
+#define LAY_EXTENSION ".lay"
+
+static inline int _string_end_with_icd(const char* str) {
+  return _string_end_with(str, ICD_EXTENSION);
+}
+
+static inline int _string_end_with_lay(const char* str) {
+  return _string_end_with(str, LAY_EXTENSION);
+}
+
 static inline int _string_with_slash(const char* str) {
   return strchr(str, '/') != NULL;
 }
 
-static inline unsigned int _find_num_icds(DIR *dir) {
-  unsigned int num_icds = 0;
+
+static inline unsigned int _find_num_suffix_match(DIR *dir, const char* suffix) {
+  unsigned int num_matches = 0;
   struct dirent *ent;
   while( (ent=readdir(dir)) != NULL ){
-    if (_string_end_with_icd(ent->d_name)) {
-      num_icds++;
+    if (_string_end_with(ent->d_name, suffix)) {
+      num_matches++;
     }
   }
   rewinddir(dir);
+  return num_matches;
+}
+
+static inline unsigned int _find_num_icds(DIR *dir) {
+  unsigned int num_icds = _find_num_suffix_match(dir, ICD_EXTENSION);
   RETURN(num_icds);
 }
 
+static inline unsigned int _find_num_lays(DIR *dir) {
+  unsigned int num_lays = _find_num_suffix_match(dir, LAY_EXTENSION);
+  RETURN(num_lays);
+}
+
+static int compare_path(const void *a, const void *b)
+{
+      return strcoll(*(const char **)a, *(const char **)b);
+}
+
 static inline unsigned int _load_icd(int num_icds, const char* lib_path) {
   unsigned int ret=0;
   debug(D_LOG, "Loading ICD '%s'", lib_path);
@@ -675,13 +685,120 @@ static void __initLayer(char * layer_path) {
     for( i = limit; i <= OCL_ICD_LAST_FUNCTION; i++) {
       ((void **)&(new_layer->dispatch))[i] = ((void **)target_dispatch)[i];
     }
+#ifdef CLLAYERINFO
+  new_layer->library_name = strdup(layer_path);
+  new_layer->layer_info_fn_ptr = clGetLayerInfo_ptr;
+#endif
   } else {
     debug(D_WARN, "Layer: %s could not be loaded", layer_path);
   }
 }
 
+static void __initSystemLayers( void ) {
+  struct stat buf;
+  cl_uint num_lays = 0;
+  int ret;
+  struct dirent *ent;
+  DIR *dir = NULL;
+  const char* opencl_layer_path=getenv("OPENCL_LAYER_PATH");
+  if (! opencl_layer_path || opencl_layer_path[0]==0) {
+    opencl_layer_path=ETC_OPENCL_LAYERS;
+    debug(D_DUMP, "OPENCL_LAYER_PATH unset or empty. Using hard-coded path '%s'", opencl_layer_path);
+  } else {
+    debug(D_DUMP, "OPENCL_LAYER_PATH set to '%s', using it", opencl_layer_path);
+  }
+  debug(D_LOG,"Reading lay list from '%s'", opencl_layer_path);
+
+  ret=stat(opencl_layer_path, &buf);
+  if (ret != 0) {
+    debug(D_WARN, "Cannot stat '%s'. Aborting", opencl_layer_path);
+    return;
+  }
+  if (!S_ISDIR(buf.st_mode)) {
+    debug(D_WARN, "'%s' is not a directory. Aborting", opencl_layer_path);
+    return;
+  }
+  debug(D_LOG,"Reading lay list from '%s'", opencl_layer_path);
+  dir = opendir(opencl_layer_path);
+  if(dir == NULL) {
+    if (errno == ENOTDIR) {
+      debug(D_DUMP, "%s is not a directory. Aborting", opencl_layer_path);
+    }
+    return;
+  }
+
+  num_lays = _find_num_lays(dir);
+  if(num_lays == 0) {
+    return;
+  }
+
+  char **dir_elems = NULL;
+  cl_uint real_num_lays = 0;
+  dir_elems = (char **)malloc(num_lays*sizeof(char *));
+  if(!dir_elems) {
+    return;
+  }
+  while( (ent=readdir(dir)) != NULL && real_num_lays < num_lays){
+    char * lib_path;
+    unsigned int lib_path_length;
+    if (!_string_end_with_lay(ent->d_name)) {
+      continue;
+    }
+    lib_path_length = strlen(opencl_layer_path) + strlen(ent->d_name) + 2;
+    lib_path = malloc(lib_path_length*sizeof(char));
+    if (!lib_path) {
+      free(lib_path);
+      continue;
+    }
+    sprintf(lib_path,"%s/%s", opencl_layer_path, ent->d_name);
+    debug(D_LOG, "Considering file '%s'", lib_path);
+    dir_elems[real_num_lays] = lib_path;
+    real_num_lays++;
+  }
+  qsort(dir_elems, real_num_lays, sizeof(char *), compare_path);
+  for(cl_uint j = 0; j < real_num_lays; j++) {
+    unsigned int lib_path_length;
+    char * err;
+    char * lib_path = dir_elems[j];
+    FILE *f = fopen(lib_path,"r");
+    free(lib_path);
+    if (f==NULL) {
+      continue;
+    }
+    fseek(f, 0, SEEK_END);
+    lib_path_length = ftell(f)+1;
+    fseek(f, 0, SEEK_SET);
+    if(lib_path_length == 1) {
+      debug(D_WARN, "File contents too short, skipping LAY");
+      fclose(f);
+      continue;
+    }
+    lib_path = malloc(lib_path_length*sizeof(char));
+    if (!lib_path) {
+      continue;
+    }
+    err = fgets(lib_path, lib_path_length*sizeof(char), f);
+    if( err == NULL ) {
+      free(lib_path);
+      debug(D_WARN, "Error while loading file contents, skipping LAY");
+      continue;
+    }
+
+    lib_path_length = strnlen(lib_path, lib_path_length);
+
+    if( lib_path[lib_path_length-1] == '\n' )
+      lib_path[lib_path_length-1] = '\0';
+
+    __initLayer(lib_path);
+    free(lib_path);
+  }
+  free(dir_elems);
+}
+
 static void __initLayers( void ) {
-  char* layers_path=getenv("OCL_ICD_LAYERS");
+  __initSystemLayers();
+
+  char* layers_path=getenv("OPENCL_LAYERS");
   if (layers_path) {
     char* layer_path = layers_path;
     char* next_layer_path = strchr(layers_path, ':');
@@ -884,15 +1001,19 @@ getDefaultPlatformID() {
 }
 
 #pragma GCC visibility pop
+#if defined(__APPLE__) || defined(__MACOSX)
+#define hidden_alias(name)
+#else
 #define hidden_alias(name) \
   typeof(name) name##_hid __attribute__ ((alias (#name), visibility("hidden")))
+#endif
 
-typedef enum {
-  CL_ICDL_OCL_VERSION=1,
-  CL_ICDL_VERSION=2,
-  CL_ICDL_NAME=3,
-  CL_ICDL_VENDOR=4,
-} cl_icdl_info;
+typedef cl_uint cl_icdl_info;
+
+#define CL_ICDL_OCL_VERSION 1
+#define CL_ICDL_VERSION     2
+#define CL_ICDL_NAME        3
+#define CL_ICDL_VENDOR      4
 
 static cl_int clGetICDLoaderInfoOCLICD(
   cl_icdl_info     param_name,
@@ -970,7 +1091,7 @@ CL_API_ENTRY void * CL_API_CALL
 clGetExtensionFunctionAddress(const char * func_name) CL_API_SUFFIX__VERSION_1_0 {
   debug_trace();
   _initClIcd();
-  if (__builtin_expect (!!_first_layer, 0))
+  if (_first_layer)
     return _first_layer->dispatch.clGetExtensionFunctionAddress(func_name);
   clGetExtensionFunctionAddress_body
 }
@@ -1012,7 +1133,7 @@ clGetPlatformIDs(cl_uint          num_entries,
                  cl_uint *        num_platforms) CL_API_SUFFIX__VERSION_1_0 {
   debug_trace();
   _initClIcd();
-  if (__builtin_expect (!!_first_layer, 0))
+  if (_first_layer)
     return _first_layer->dispatch.clGetPlatformIDs(num_entries, platforms, num_platforms);
   clGetPlatformIDs_body
 }
@@ -1092,7 +1213,7 @@ clCreateContext(const cl_context_properties *  properties,
                 cl_int *                       errcode_ret) CL_API_SUFFIX__VERSION_1_0 {
   debug_trace();
   _initClIcd();
-  if (__builtin_expect (!!_first_layer, 0))
+  if (_first_layer)
     return _first_layer->dispatch.clCreateContext(properties,
                                                   num_devices,
                                                   devices,
@@ -1149,7 +1270,7 @@ clCreateContextFromType(const cl_context_properties *  properties,
                         cl_int *                       errcode_ret) CL_API_SUFFIX__VERSION_1_0 {
   debug_trace();
   _initClIcd();
-  if (__builtin_expect (!!_first_layer, 0))
+  if (_first_layer)
     return _first_layer->dispatch.clCreateContextFromType(properties,
                                                           device_type,
                                                           pfn_notify,
@@ -1197,7 +1318,7 @@ clGetGLContextInfoKHR(const cl_context_properties *  properties,
                       size_t *                       param_value_size_ret) CL_API_SUFFIX__VERSION_1_0 {
   debug_trace();
   _initClIcd();
-  if (__builtin_expect (!!_first_layer, 0))
+  if (_first_layer)
     return _first_layer->dispatch.clGetGLContextInfoKHR(properties,
                                                         param_name,
                                                         param_value_size,
@@ -1225,7 +1346,7 @@ CL_API_ENTRY cl_int CL_API_CALL
 clWaitForEvents(cl_uint              num_events,
                 const cl_event *     event_list) CL_API_SUFFIX__VERSION_1_0 {
   debug_trace();
-  if (__builtin_expect (!!_first_layer, 0))
+  if (_first_layer)
     return _first_layer->dispatch.clWaitForEvents(num_events,
                                                   event_list);
   clWaitForEvents_body
@@ -1243,7 +1364,7 @@ clUnloadCompiler_disp(void) {
 CL_API_ENTRY cl_int CL_API_CALL
 clUnloadCompiler(void) CL_API_SUFFIX__VERSION_1_0 {
   debug_trace();
-  if (__builtin_expect (!!_first_layer, 0))
+  if (_first_layer)
     return _first_layer->dispatch.clUnloadCompiler();
   clUnloadCompiler_body
 }
